// Krb5Plugin.cpp : Initialisierungsroutinen
#include <shlwapi.h>
#include "Krb5Plugin.h"

#define TKT_LIFETIME     30           /* 30 seconds */
#define USE_MEMORY_CC                 /* use in-memory ccache */

HINSTANCE hInst;

// Ldt String aus Ressource und dupliziert ihn auf den lokalen Windows-Heap
// Die Stringpuffer-Freigabe ist unklar!
LPTSTR LoadStr(UINT id) {
 TCHAR buf[256];
 LoadString(hInst,id,buf,elemof(buf));
 return StrDup(buf);
}

KRB5PLUGIN_API BOOL UserLogin(LPTSTR Username, LPTSTR Password, pGinaInfo *settingsInfo) {
 krb5_error_code	err;
 krb5_context		context;
 krb5_ccache		ccache;
 krb5_principal		user_principal;
 krb5_get_init_creds_opt gic_options;
 krb5_creds		credentials;
 char			*realm = NULL;
 bool			have_realm = false,
			have_user_principal = false,
			have_credentials = false,
			ret = false;
    // bypass local administrator
 if (!lstrcmpi(Username,T("administrator"))) {
       // dont return true, only use local account
  settingsInfo->errorString = LoadStr(102);	// "Check Admin-Account locally!"
  return false;
 }
	// init kerberos context
 err = krb5_init_context(&context);
 if (err) {
  settingsInfo->errorString = LoadStr(103);	// "Fehler beim Initialisieren des Kerberos-Kontextes."
  goto cleanup;
 }
 err = krb5_get_default_realm(context,&realm);
 if (err) {
  settingsInfo->errorString = LoadStr(104);	// "Fehler beim Ermitteln der Standard-Kerberoszelle."
  goto cleanup;
 }
 have_realm = true;

#ifdef USE_MEMORY_CC
 char ccache_name[MAX_PATH];
	// Note, Kerberos use on Windows the API:krb5cc Ticket Cache (shared memory)
 lstrcpynA(ccache_name,"MEMORY:",elemof(ccache_name));
 GetTempFileNameA(".","krb",0,ccache_name+7);
// tmpnam_s(ccache_name+7,sizeof(ccache_name)-7);
 err = krb5_cc_resolve(context, ccache_name, &ccache);
#else
    err = krb5_cc_default(context, &ccache);
#endif
 if (err) {
  settingsInfo->errorString = LoadStr(105);	// "Fehler bei Auswahl des Credential Caches."
  goto cleanup;
 }
 
 char UsernameA[256];	// UTF-8-Version vom Nutzerkennzeichen
 WideCharToMultiByte(CP_UTF8,0,Username,-1,UsernameA,elemof(UsernameA),NULL,NULL);

 err = krb5_parse_name(context,UsernameA,&user_principal);
 if (err) {
  settingsInfo->errorString = LoadStr(106);	// "Fehler beim Ermitteln des Principals."
  goto cleanup;
 }
 have_user_principal = 1;

 err = krb5_cc_initialize(context, ccache, user_principal);
 if (err) {
  settingsInfo->errorString = LoadStr(107);	// "Fehler beim Initialisieren des Credential Caches."
  goto cleanup;
 }

 krb5_get_init_creds_opt_init(&gic_options);
 krb5_get_init_creds_opt_set_tkt_life(&gic_options, TKT_LIFETIME);

 RtlZeroMemory(&credentials,sizeof(credentials));

 char PasswordA[256];	// UTF-8-Version vom Passwort (modern?)
 WideCharToMultiByte(CP_UTF8,0,Password,-1,PasswordA,elemof(PasswordA),NULL,NULL);

 err = krb5_get_init_creds_password(context, &credentials,
   user_principal, PasswordA, 0, 0, 0, 0, &gic_options);

 switch (err) {
  case 0:
        /* Success. */
		// This user can change their password
   settingsInfo->allowPassChange = false;
        // This users description
   settingsInfo->userDescription = LoadStr(108);	// "User, created by Krb5Plugin"
        // This users full name
   settingsInfo->userFullName = StrDup(Username);
        // This user should be in the Administrator group
   settingsInfo->isAdmin = false;
        // Return Message
   settingsInfo->errorString = LoadStr(109);	// "Password o.k."
   have_credentials = 1;
   ret = true;
  break;
  case KRB5KDC_ERR_PREAUTH_FAILED:
  case KRB5KRB_AP_ERR_BAD_INTEGRITY:
  case KRB5KDC_ERR_C_PRINCIPAL_UNKNOWN:
        /* Bad username or password, unknown user etc */
   settingsInfo->errorString = LoadStr(110);	// "Nutzername oder Passwort inkorrekt!"
  break;
  default:
        /* Some other sort of failure. */
   settingsInfo->errorString = LoadStr(111);	// "Unbekannter Fehler bei der Authentifizierung!"
 }
    
cleanup:
 if (have_user_principal) krb5_free_principal(context, user_principal);
 if (have_credentials) krb5_free_cred_contents(context, &credentials);
 if (realm) krb5_free_default_realm(context,realm);
 if (context) krb5_free_context(context);
    // return result
 return ret;
}

KRB5PLUGIN_API BOOL ChangeUserPassword(LPTSTR Username,LPTSTR OldPassword,LPTSTR NewPassword) {
	// return always false, its not allowed
 return false;
}

KRB5PLUGIN_API LPCTSTR AboutPlugin(void) {
 return LoadStr(100);			// "Kerberos5 Authentifizierungs-Plugin."
}

KRB5PLUGIN_API void ChangePluginSettings(void) {
 DialogBox(hInst,MAKEINTRESOURCE(101),0,(DLGPROC)ConfigDlgProc);
}

KRB5PLUGIN_API void LoginHook(pGinaInfo *settingsInfo) {
	// get kerberos ticket
	/*FILE            *debugFile;
	debugFile = fopen("C:\\tmp\\krb5plugin.log", "a");
	fprintf(debugFile, "call LoginHook: %s.\n\n", convert_to_char(settingsInfo->userDescription));
	fclose(debugFile); */
}

KRB5PLUGIN_API void LogoutHook(pGinaInfo *settingsInfo) {
}

KRB5PLUGIN_API BOOL IsRequired(void) {
 return false;
}

EXTERN_C BOOL APIENTRY _DllMainCRTStartup(HMODULE hModule, DWORD reason, LPVOID) {
 switch (reason) {
  case DLL_PROCESS_ATTACH: {
   hInst=(HINSTANCE)hModule;
   DisableThreadLibraryCalls(hModule);
  }break;
 }
 return TRUE;
}
